import fnmatch
import re
import sys
import warnings
from typing import Union

from boa_restrictor.common.rule import Rule
from boa_restrictor.rules import BOA_RESTRICTOR_RULES

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
from pathlib import Path


def load_configuration(*, file_path: Union[Path, str] = "pyproject.toml") -> dict:
    """
    Load linter configuration from pyproject.toml file.
    """
    file_path = Path.cwd() / file_path
    try:
        with open(file_path, "rb") as f:
            data = tomllib.load(f)
    except FileNotFoundError:
        return {}

    try:
        return data["tool"]["boa-restrictor"]
    except KeyError:
        return {}


def is_rule_excluded(*, rule_class: type[Rule], excluded_rules: list) -> bool:
    """
    Check if the given rule is in the exclusion list.
    """
    # Generate set of valid rules
    valid_rules = (rule_class.RULE_ID for rule_class in BOA_RESTRICTOR_RULES)

    # Check if the given rule is valid
    for invalid_configured_rule in [rule_id for rule_id in excluded_rules if rule_id not in valid_rules]:
        warnings.warn(
            f'Boa Restrictor: Invalid rule "{invalid_configured_rule}" in configuration detected.',
            category=UserWarning,
            stacklevel=2,
        )
        return False

    # Check if the given rule is in the exclusion list
    return rule_class.RULE_ID in excluded_rules


def is_rule_excluded_per_file(
    *, filename: str, rule_class: type[Rule], per_file_excluded_rules: dict[str, list]
) -> bool:
    """
    Check if the given rule is in the per-file-exclusion list.
    """
    # Iterate per-file rule exclusions
    for file_path_pattern in per_file_excluded_rules.keys():
        # If the filename matches the pattern...
        if re.search(fnmatch.translate(file_path_pattern), filename):
            # Skip linters, which have been excluded for this file path pattern
            if is_rule_excluded(rule_class=rule_class, excluded_rules=per_file_excluded_rules[file_path_pattern]):
                return True
    return False



from unittest import mock

import pytest

from boa_restrictor.common.rule import Rule


@mock.patch.object(Rule, "check")
def test_run_check(mocked_check):
    Rule.run_check(filename="my/file.py", source_tree="Python!")

    mocked_check.assert_called_with()
    mocked_check.assert_called_once()


def test_init_variables_set():
    rule = Rule(filename="my/file.py", source_tree="Python!")

    assert rule.filename == "my/file.py"
    assert rule.source_tree == "Python!"


def test_check_not_implemented():
    with pytest.raises(NotImplementedError):
        Rule.run_check(filename="my/file.py", source_tree="Python!")



import ast

import pytest

from boa_restrictor.cli.utils import parse_source_code_or_fail
from boa_restrictor.exceptions.syntax_errors import BoaRestrictorParsingError


def test_parse_source_code_or_fail_valid_code():
    result = parse_source_code_or_fail(filename="my_file.py", source_code="a = 1")

    assert isinstance(result, ast.Module)


def test_parse_source_code_or_fail_invalid_code():
    with pytest.raises(BoaRestrictorParsingError, match='Source code of file "my_file.py" contains syntax errors.'):
        parse_source_code_or_fail(filename="my_file.py", source_code="""f\"asf""")



import pytest

from boa_restrictor.exceptions.syntax_errors import BoaRestrictorParsingError


def test_parsing_error():
    exception = BoaRestrictorParsingError(filename="my_file")

    with pytest.raises(BoaRestrictorParsingError, match='Source code of file "my_file" contains syntax errors.'):
        raise exception



from boa_restrictor.rules.abstract_class_inherits_from_abc import AbstractClassesInheritFromAbcRule
from boa_restrictor.rules.asterisk_required import AsteriskRequiredRule
from boa_restrictor.rules.dataclass_kw_only import DataclassWithKwargsOnlyRule
from boa_restrictor.rules.global_import_datetime import GlobalImportDatetimeRule
from boa_restrictor.rules.return_type_hints import ReturnStatementRequiresTypeHintRule
from boa_restrictor.rules.service_class_only_one_public import ServiceClassHasOnlyOnePublicMethodRule

BOA_RESTRICTOR_RULES = (
    AsteriskRequiredRule,
    ReturnStatementRequiresTypeHintRule,
    GlobalImportDatetimeRule,
    DataclassWithKwargsOnlyRule,
    ServiceClassHasOnlyOnePublicMethodRule,
    AbstractClassesInheritFromAbcRule,
)



from pathlib import Path

from boa_restrictor.rules import BOA_RESTRICTOR_RULES


def test_boa_restrictor_rules_constant_not_missing_rules():
    """
    This test is a check to ensure we don't forget to register new rules.
    """
    number_of_rule_files = 0
    for file in (Path(__file__).resolve().parent.parent.parent / "boa_restrictor/rules").iterdir():
        if file.suffix == ".py" and file.name != "__init__.py":
            number_of_rule_files += 1

    assert len(BOA_RESTRICTOR_RULES) == number_of_rule_files



import os
import sys
import warnings

from boa_restrictor.rules import AsteriskRequiredRule

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
from unittest import mock

from boa_restrictor.cli.configuration import is_rule_excluded, is_rule_excluded_per_file, load_configuration


@mock.patch.object(tomllib, "load", return_value={"tool": {"boa-restrictor": {"exclude": ["PBR001"]}}})
def test_load_configuration_happy_path(mocked_load):
    data = load_configuration(file_path=os.path.abspath(sys.argv[0]))

    mocked_load.assert_called_once()
    assert data == {"exclude": ["PBR001"]}


@mock.patch.object(
    tomllib, "load", return_value={"tool": {"boa-restrictor": {"per-file-excludes": {"*/test/*": ["PBR001"]}}}}
)
def test_load_configuration_per_file_excludes(mocked_load):
    data = load_configuration(file_path=os.path.abspath(sys.argv[0]))

    mocked_load.assert_called_once()
    assert data == {"per-file-excludes": {"*/test/*": ["PBR001"]}}


@mock.patch.object(tomllib, "load")
def test_load_configuration_invalid_file(mocked_load):
    data = load_configuration(file_path="invalid_file.toml")

    mocked_load.assert_not_called()
    assert data == {}


@mock.patch.object(tomllib, "load", return_value={"tool": {"other_linter": True}})
def test_load_configuration_key_missing(mocked_load):
    data = load_configuration(file_path=os.path.abspath(sys.argv[0]))

    mocked_load.assert_called_once()
    assert data == {}


def test_is_rule_excluded_is_excluded():
    assert is_rule_excluded(rule_class=AsteriskRequiredRule, excluded_rules=["PBR001"]) is True


def test_is_rule_excluded_is_not_excluded():
    assert is_rule_excluded(rule_class=AsteriskRequiredRule, excluded_rules=["PBR002"]) is False


@mock.patch.object(warnings, "warn")
def test_is_rule_excluded_invalid_rule(mocked_warn):
    assert is_rule_excluded(rule_class=AsteriskRequiredRule, excluded_rules=["PBR999"]) is False
    mocked_warn.assert_called_once()


def test_is_rule_excluded_per_file_is_excluded():
    assert (
        is_rule_excluded_per_file(
            filename="tests/test_history.py",
            rule_class=AsteriskRequiredRule,
            per_file_excluded_rules={"*.py": ["PBR001"]},
        )
        is True
    )


def test_is_rule_excluded_per_file_is_not_excluded():
    assert (
        is_rule_excluded_per_file(
            filename="tests/test_history.py",
            rule_class=AsteriskRequiredRule,
            per_file_excluded_rules={"*.py": ["PBR002"]},
        )
        is False
    )


def test_is_rule_excluded_per_file_file_not_matched():
    assert (
        is_rule_excluded_per_file(
            filename="pyproject.toml",
            rule_class=AsteriskRequiredRule,
            per_file_excluded_rules={"*.py": ["PBR002"]},
        )
        is False
    )



import ast

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.service_class_only_one_public import ServiceClassHasOnlyOnePublicMethodRule


def test_functions_named_service_are_ignored():
    source_tree = ast.parse("""def myFunctionService():
            pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_classes_not_named_service_are_ignored():
    source_tree = ast.parse("""class BeerHeuristics:
            pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_service_class_missing_public_method():
    source_tree = ast.parse("""class MyService:
            pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=ServiceClassHasOnlyOnePublicMethodRule.RULE_ID,
        rule_label=ServiceClassHasOnlyOnePublicMethodRule.RULE_LABEL,
        identifier="MyService",
    )


def test_service_class_with_public_process_method():
    source_tree = ast.parse("""class MyService:
            def process():
                pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_service_class_with_protected_method():
    source_tree = ast.parse("""class MyService:
            def _process():
                pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=ServiceClassHasOnlyOnePublicMethodRule.RULE_ID,
        rule_label=ServiceClassHasOnlyOnePublicMethodRule.RULE_LABEL,
        identifier="MyService",
    )


def test_service_class_with_multiple_public_methods():
    source_tree = ast.parse("""class MyService:
            def durp():
                pass

            def process():
                pass""")

    occurrences = ServiceClassHasOnlyOnePublicMethodRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=ServiceClassHasOnlyOnePublicMethodRule.RULE_ID,
        rule_label=ServiceClassHasOnlyOnePublicMethodRule.RULE_LABEL,
        identifier="MyService",
    )



import ast

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.dataclass_kw_only import DataclassWithKwargsOnlyRule


def test_dataclass_kw_only_set():
    source_tree = ast.parse("""import dataclass
@dataclasses.dataclass(kw_only=True)
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_other_class_decorator():
    source_tree = ast.parse("""
@abc.ABCMeta.register
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_dataclass_kw_only_missing():
    source_tree = ast.parse("""import dataclass
@dataclasses.dataclass
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=3,
        rule_id=DataclassWithKwargsOnlyRule.RULE_ID,
        rule_label=DataclassWithKwargsOnlyRule.RULE_LABEL,
        identifier=None,
    )


def test_dataclass_kw_only_set_but_false():
    source_tree = ast.parse("""import dataclass
@dataclasses.dataclass(kw_only=False)
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=3,
        rule_id=DataclassWithKwargsOnlyRule.RULE_ID,
        rule_label=DataclassWithKwargsOnlyRule.RULE_LABEL,
        identifier=None,
    )


def test_nested_import_kwargs_set():
    source_tree = ast.parse("""from dataclasses import dataclass
@dataclass(kw_only=True)
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_nested_import_kwargs_missing():
    source_tree = ast.parse("""from dataclasses import dataclass
@dataclass
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=3,
        rule_id=DataclassWithKwargsOnlyRule.RULE_ID,
        rule_label=DataclassWithKwargsOnlyRule.RULE_LABEL,
        identifier=None,
    )


def test_nested_import_kwargs_set_but_false():
    source_tree = ast.parse("""from dataclasses import dataclass
@dataclass(kw_only=False)
class MyDataclass:
    pass""")

    occurrences = DataclassWithKwargsOnlyRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=3,
        rule_id=DataclassWithKwargsOnlyRule.RULE_ID,
        rule_label=DataclassWithKwargsOnlyRule.RULE_LABEL,
        identifier=None,
    )
